----------------------------------------------------------------------------------
-- Company: 
-- Engineer: 
-- 
-- Create Date:    22:16:35 08/24/2016 
-- Design Name: 
-- Module Name:    hdmi_encoder - Behavioral 
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description: 
--
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;

-- Uncomment the following library declaration if using
-- arithmetic functions with Signed or Unsigned values
--use IEEE.NUMERIC_STD.ALL;

-- Uncomment the following library declaration if instantiating
-- any Xilinx primitives in this code.
library UNISIM;
use UNISIM.VComponents.all;

entity hdmi_encoder is
	PORT(

		-- pixel and TMDS clock
		pixel_clock     : IN std_logic;
		clock_5x     	 : IN std_logic;
	
		-- input: XGA signal 
		red, green, blue: IN std_logic_vector(7 downto 0);
		blank           : IN std_logic;
		hsync           : IN std_logic;
		vsync           : IN std_logic;
		
		-- HDMI output
		hdmi_p : OUT std_logic_vector(3 downto 0);
		hdmi_n : OUT std_logic_vector(3 downto 0)

  );
end hdmi_encoder;

architecture Behavioral of hdmi_encoder is
	COMPONENT tmds_encoder
		PORT( 
			clk: in std_logic;
			VD: in std_logic_vector (7 downto 0);	-- video data (red, green or blue)
			vsync: in std_logic;
			hsync: in std_logic;
			VDE: in std_logic;  			-- video data enable, to choose between CD (when VDE=0) and VD (when VDE=1)
			TMDS: out std_logic_vector (9 downto 0)
		);
	END COMPONENT;
 
   type a_symbols     is array (0 to 3) of std_logic_vector(9 downto 0);
   type a_output_bits is array (0 to 3) of std_logic_vector(1 downto 0);

   signal symbols        : a_symbols     := (others => (others => '0'));
   signal high_speed_sr  : a_symbols     := (others => (others => '0'));
   signal output_bits    : a_output_bits := (others => (others => '0'));
   
   -- Controlling when the transfers into the high speed domain occur
   signal latch_high_speed : std_logic_vector(4 downto 0) := "00001";
   
   -- The signals from the DDR outputs to the output buffers
   signal serial_outputs : std_logic_vector(3 downto 0);
	
begin

	encoder_red: tmds_encoder 
	PORT MAP (
		clk => pixel_clock,
		VD => red,
		vsync => '0',
		hsync => '0',
		VDE => not blank,
		TMDS => symbols (2)
	);

	encoder_green: tmds_encoder 
	PORT MAP (
		clk => pixel_clock,
		VD => green,
		vsync => '0',
		hsync => '0',
		VDE => not blank,
		TMDS => symbols (1)
	);

	encoder_blue: tmds_encoder 
	PORT MAP (
		clk => pixel_clock,
		VD => blue,
		vsync => vsync,
		hsync => hsync,
		VDE => not blank,
		TMDS => symbols (0)
	);

   symbols(3) <= "0000011111"; -- the clock channel symbol is static 
	
	process(clock_5x)
   begin
      ---------------------------------------------------------------
      -- Now take the 10-bit words and take it into the high-speed
      -- clock domain once every five cycles. 
      -- 
      -- Then send out two bits every clock cycle using DDR output
      -- registers.
      ---------------------------------------------------------------   
      if rising_edge(clock_5x) then
         for i in 0 to 3 loop
            output_bits(i)  <= high_speed_sr(i)(1 downto 0);
            if latch_high_speed(0) = '1' then
               high_speed_sr(i) <= symbols(i);
            else
               high_speed_sr(i) <= "00" & high_speed_sr(i)(9 downto 2);
            end if;
         end loop;
         latch_high_speed <= latch_high_speed(0) & latch_high_speed(4 downto 1);
      end if;
   end process;

	g1:   for i in 0 to 3 generate
   --------------------------------------------------------
   -- Convert the TMDS codes into a serial stream, two bits 
   -- at a time using a DDR register
   --------------------------------------------------------
      to_serial: ODDR2
         generic map(DDR_ALIGNMENT => "C0", INIT => '0', SRTYPE => "ASYNC") 
         port map (C0 => clock_5x,  C1 => not clock_5x, CE => '1', R => '0', S => '0',
                   D0 => output_bits(i)(0), D1 => output_bits(i)(1), Q => serial_outputs(i));
      OBUFDS_c0  : OBUFDS port map ( O  => hdmi_p(i), OB => hdmi_n(i), I => serial_outputs(i));
   end generate;

end Behavioral;

